import torch
import torch_scatter

class GarnetExchange(torch.nn.Module):
    def __init__(self, 
                 in_channels : int,
                 aggregators : int,
                 sumw : bool = False,
                 ):
        super(GarnetExchange, self).__init__()

        self.lin = torch.nn.Linear(in_channels, aggregators)
        self.aggregators = aggregators
        self.sumw = sumw
        
    def forward(self, x, batch):

        w = self.lin(x)
        w = torch.exp(- torch.abs(w))
        if self.sumw:
            w = w / (torch.sum(w, dim=1, keepdim=True) + 1e-4)
        
            

        return x


        
    

from torch_geometric.data import Data

# Example node features (x) for three graphs
x1 = torch.tensor([[1, 2], [3, 4], [5, 6], [7, 8]], dtype=torch.float)  # Graph 1
x2 = torch.tensor([[9, 10], [11, 12], [13, 14]], dtype=torch.float)    # Graph 2
x3 = torch.tensor([[15, 16], [17, 18], [19, 20], [21, 22], [23, 24]], dtype=torch.float)  # Graph 3

# Combine node features
x_combined = torch.cat([x1, x2, x3], dim=0)

# Create the batch vector
batch = torch.tensor([0]*4 + [1]*3 + [2]*5)

# Create a Data object for PyTorch Geometric
data = Data(x=x_combined, batch=batch)


gn = GarnetExchange(2, 2, sumw=True)
print(gn(data.x, data.batch))