from argparse import ArgumentParser
from smooth_animate import animate_smoothed_data
import pickle
import numpy as np
from generator import Generator, Parameters
from reconstruction import Reconstruction
import torch

def apply_model_to_parameters(model, parameters, particles = [['gamma',0.22]], n_events_per_var=2000):
    # create a photon dataset based on the parameters using the generator class
    
    
    # get the data from the generator
    ds = gen.create_torch_dataset(random_scaler=0.)
    # apply the model to the data
    res,_ = model.apply_model_in_batches(ds, batch_size=1000)
    res = ds.unnormalise_target(res.to('cuda')).detach().cpu().numpy()
    truth = ds.unnormalise_target(torch.tensor(ds.target_array).to('cuda')).detach().cpu().numpy()
    return res, truth

if __name__ == "__main__":
    # read in the pickle evolution file from execute.py from command line
    parser = ArgumentParser(description="Plot the evolution of the data")
    parser.add_argument("file", help="The pickle file to be plotted")
    parser.add_argument("--output", help="The output file name", default="evolution")
    parser.add_argument("--n", help="The number of steps to be plotted", default=100, type=int)
    args = parser.parse_args()
    file = args.file
    with open(file, "rb") as f:
        data = pickle.load(f)

    data_to_plot = []
    labels = [] #same
    all_data = []
    for i,_e in enumerate(data):
        # save progress to output log file just a counter i/all
        with open(args.output + 'progress.log', 'a') as f:
            f.write(f"{i}/{len(data)}\n")
        
        if i > args.n:
            break
        e = _e[0]
        #get the model
        state_dict = e['reco_model']
        means = e['means']
        stds = e['stds']
        e.pop('reco_model')
        e.pop('means')
        e.pop('stds')
        #get the detector parameters
        dp = e

        print(dp)
        ress = []
        labels = []
        bin_centers = []
        this_data = []

        for particle in [[['gamma',0.22]], [['pi+',2.11]]]:

            gen = Generator(np.array(len(dp)*[1e-9]), 
                    Parameters(parameters = dp),
                    n_vars = 1,
                    n_events_per_var = 2000,
                    particles=particle)#second entry only important for the surrogate, ignored for reco
    
            #recreate the pytorch model from the state dictionary
            model = Reconstruction(gen.n_parameters, 
                               gen.n_parameters_per_sensor * gen.n_sensors, 
                               gen.n_target_parameters)
            model.load_state_dict(state_dict)
            
            res, truth = apply_model_to_parameters(model, dp, means, stds)
            resolution = (res-truth)/truth
    
            #make a histogram of the resolution between -1 and 1 with 21 bins
            res, edges = np.histogram(resolution, bins=np.linspace(-1,1,21))
            # get the bin centers using linspace again
            bin_centers = np.linspace(-1,1,21)[:-1] + (np.linspace(-1,1,21)[1]-np.linspace(-1,1,21)[0])/2


            ress.append(res)
            labels.append(particle)
            this_data.append([particle, resolution])
        
        data_to_plot.append(ress)
        all_data.append(this_data)
  
    print('creating ', len(data_to_plot), ' frames')
    #save all_data this pickle
    with open(args.output+ 'all_data.pickle', 'wb') as f:
        pickle.dump(all_data, f)
    #plot the data
    data_to_plot = np.array(data_to_plot)
    bin_centers = np.array(bin_centers)
    print(data_to_plot.shape, len(labels), len(bin_centers))
    animate_smoothed_data(data_to_plot, labels, bin_centers, x_label="Resolution", y_label="Counts",
                           interval=30, outfile=f"{args.output}resolution_evolution.gif")


