# contains the reconstrction module
# neural network model for reconstruction of energy from detector depostits and parameters
# contains also the training loop
# the dataset is provided by the Generator as a pandas dataframe
# Data loader etc are in here, too

import torch
import torch.utils.data
from torch.utils.data import DataLoader
import json

#a simple reconstruction model, just feed-forward for now

class Reconstruction(torch.nn.Module):
    def __init__(self, n_detector_parameters, n_input_parameters, n_target_parameters):
        super(Reconstruction, self).__init__()

        with open('config.json', 'r') as config_file:
             config = json.load(config_file)
             
        self.n_parameters = n_detector_parameters

        self.preprocess = torch.nn.Sequential(
            torch.nn.Linear(n_detector_parameters, 100),
            torch.nn.ELU(),
            torch.nn.Linear(100,100),
            torch.nn.ELU(),
            torch.nn.Linear(100, n_input_parameters),
            torch.nn.ReLU()
        )
        # take into account that 
        self.layers = torch.nn.Sequential(
            #torch.nn.BatchNorm1d(n_detector_parameters + n_input_parameters),
            torch.nn.Linear(n_detector_parameters + n_input_parameters, 100),
            torch.nn.ELU(),
            torch.nn.Linear(100, 100),
            torch.nn.ELU(),
            #torch.nn.BatchNorm1d(100),
            torch.nn.Linear(100, 100),
            torch.nn.ELU(),
            #torch.nn.BatchNorm1d(100),
            torch.nn.Linear(100, n_target_parameters),
        )

        # some placeholders for a simpler training loop
        self.optimizer = torch.optim.Adam(self.parameters(), lr=0.0001)
        self.device = torch.device(config["device"]) #cuda

    def forward(self, detector_parameters, x):
        #concatenate the detector parameters and the input
        scaling_factors = self.preprocess(detector_parameters)
        x_scaled = torch.multiply(x, scaling_factors)
        concatenated_input = torch.cat([detector_parameters, x_scaled], dim=1)
        return self.layers(concatenated_input)
    
    def loss(self, y_pred, y):
        #to make it more stable filter nans or infs based on the predicted first from y, then from pedicted
        y = torch.where(torch.isnan(y_pred), torch.zeros_like(y)+1., y)
        y = torch.where(torch.isinf(y_pred), torch.zeros_like(y)+1., y)
        y_pred = torch.where(torch.isnan(y_pred), torch.zeros_like(y_pred), y_pred)
        y_pred = torch.where(torch.isinf(y_pred), torch.zeros_like(y_pred), y_pred)

        return ((y_pred-y)**2 / (torch.abs(y)+1.)).mean() #calo loss with 1/sqrt(E) weighting
        #return torch.nn.MSELoss()(y_pred, y)**(1/2) #TBI properly
    
    def train_model(self, dataset, batch_size, n_epochs, lr):

        train_loader = DataLoader(dataset,  batch_size=batch_size, shuffle=True)
        # set the optimizer
        for param_group in self.optimizer.param_groups:
            param_group['lr'] = lr
        self.to(self.device)
        self.train()
        for epoch in range(n_epochs):
            for batch_idx, (detector_parameters, x, y) in enumerate(train_loader):
                detector_parameters = detector_parameters.to(self.device)
                x = x.to(self.device)
                y = y.to(self.device)
                y_pred = self(detector_parameters, x)
                loss = self.loss( dataset.unnormalise_target(y_pred), dataset.unnormalise_target(y))
                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()

            print('Reco Epoch: {} \tLoss: {:.8f}'.format(
                        epoch, loss.item()))
        self.eval()


    def apply_model_in_batches(self, dataset, batch_size):
        # apply the model in batches
        # this is necessary because the model is too large to apply it to the whole dataset at once
        # the model is applied to the dataset in batches and the results are concatenated
        # the batch size is a hyperparameter
        self.to(self.device)
        self.eval()
        # create a dataloader for the dataset
        data_loader = DataLoader(dataset, batch_size=batch_size, shuffle=False)
        # create a tensor to store the results
        results = None
        mean_loss = 0.
        # loop over the batches
        for batch_idx, (detector_parameters, x, y) in enumerate(data_loader):
            detector_parameters = detector_parameters.to(self.device)
            x = x.to(self.device)
            y = y.to(self.device)
            # apply the model
            y_pred = self(detector_parameters, x)
            # calculate the loss
            loss = self.loss( dataset.unnormalise_target(y_pred), dataset.unnormalise_target(y))
            mean_loss += loss.item()
            if results is None:
                results = torch.zeros(len(dataset), y_pred.shape[1])
            # store the results
            results[batch_idx * batch_size : (batch_idx + 1) * batch_size] = y_pred
        mean_loss /= len(data_loader)

        return results, mean_loss